@@ -2,16 +2,18 @@ package machine
 
 import (
 	"context"
-	"sigs.k8s.io/yaml"
 	"testing"
 	"time"
 
+	"reflect"
+
 	bmoapis "github.com/metal3-io/baremetal-operator/pkg/apis"
 	bmh "github.com/metal3-io/baremetal-operator/pkg/apis/metal3/v1alpha1"
 	"github.com/metal3-io/baremetal-operator/pkg/utils"
 	bmv1alpha1 "github.com/openshift/cluster-api-provider-baremetal/pkg/apis/baremetal/v1alpha1"
 	machinev1beta1 "github.com/openshift/machine-api-operator/pkg/apis/machine/v1beta1"
 	machineapierrors "github.com/openshift/machine-api-operator/pkg/controller/machine"
+	"github.com/stretchr/testify/assert"
 	corev1 "k8s.io/api/core/v1"
 	"k8s.io/apimachinery/pkg/api/errors"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
@@ -19,6 +21,7 @@ import (
 	"k8s.io/apimachinery/pkg/types"
 	"sigs.k8s.io/controller-runtime/pkg/client"
 	fakeclient "sigs.k8s.io/controller-runtime/pkg/client/fake"
+	"sigs.k8s.io/yaml"
 )
 
 const (
@@ -46,12 +49,22 @@ func TestChooseHost(t *testing.T) {
 				APIVersion: machinev1beta1.SchemeGroupVersion.String(),
 			},
 		},
+		Status: bmh.BareMetalHostStatus{
+			Provisioning: bmh.ProvisionStatus{
+				State: bmh.StateReady,
+			},
+		},
 	}
 	host2 := bmh.BareMetalHost{
 		ObjectMeta: metav1.ObjectMeta{
 			Name:      "host2",
 			Namespace: "myns",
 		},
+		Status: bmh.BareMetalHostStatus{
+			Provisioning: bmh.ProvisionStatus{
+				State: bmh.StateReady,
+			},
+		},
 	}
 	host3 := bmh.BareMetalHost{
 		ObjectMeta: metav1.ObjectMeta{
@@ -66,12 +79,22 @@ func TestChooseHost(t *testing.T) {
 				APIVersion: machinev1beta1.SchemeGroupVersion.String(),
 			},
 		},
+		Status: bmh.BareMetalHostStatus{
+			Provisioning: bmh.ProvisionStatus{
+				State: bmh.StateReady,
+			},
+		},
 	}
 	host4 := bmh.BareMetalHost{
 		ObjectMeta: metav1.ObjectMeta{
 			Name:      "host4",
 			Namespace: "someotherns",
 		},
+		Status: bmh.BareMetalHostStatus{
+			Provisioning: bmh.ProvisionStatus{
+				State: bmh.StateReady,
+			},
+		},
 	}
 	discoveredHost := bmh.BareMetalHost{
 		ObjectMeta: metav1.ObjectMeta{
@@ -80,14 +103,56 @@ func TestChooseHost(t *testing.T) {
 		},
 		Status: bmh.BareMetalHostStatus{
 			ErrorMessage: "this host is discovered and not usable",
+			Provisioning: bmh.ProvisionStatus{
+				State: bmh.StateRegistrationError,
+			},
 		},
 	}
-	host_with_label := bmh.BareMetalHost{
+	hostWithLabel := bmh.BareMetalHost{
 		ObjectMeta: metav1.ObjectMeta{
 			Name:      "host_with_label",
 			Namespace: "myns",
 			Labels:    map[string]string{"key1": "value1"},
 		},
+		Status: bmh.BareMetalHostStatus{
+			Provisioning: bmh.ProvisionStatus{
+				State: bmh.StateReady,
+			},
+		},
+	}
+	externallyProvisionedHost := bmh.BareMetalHost{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      "externally-provisioned-host",
+			Namespace: "myns",
+		},
+		Spec: bmh.BareMetalHostSpec{
+			ExternallyProvisioned: true,
+		},
+		Status: bmh.BareMetalHostStatus{
+			Provisioning: bmh.ProvisionStatus{
+				State: bmh.StateExternallyProvisioned,
+			},
+		},
+	}
+	externallyProvisionedAndConsumedHost := bmh.BareMetalHost{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      "externally-provisioned-and-consumed-host",
+			Namespace: "myns",
+		},
+		Spec: bmh.BareMetalHostSpec{
+			ExternallyProvisioned: true,
+			ConsumerRef: &corev1.ObjectReference{
+				Name:       "machine1",
+				Namespace:  "myns",
+				Kind:       "Machine",
+				APIVersion: machinev1beta1.SchemeGroupVersion.String(),
+			},
+		},
+		Status: bmh.BareMetalHostStatus{
+			Provisioning: bmh.ProvisionStatus{
+				State: bmh.StateExternallyProvisioned,
+			},
+		},
 	}
 
 	config, providerSpec := newConfig(t, "", map[string]string{}, []bmv1alpha1.HostSelectorRequirement{})
@@ -111,13 +176,14 @@ func TestChooseHost(t *testing.T) {
 		})
 
 	testCases := []struct {
+		Scenario         string
 		Machine          machinev1beta1.Machine
 		Hosts            []runtime.Object
 		ExpectedHostName string
 		Config           *bmv1alpha1.BareMetalMachineProviderSpec
 	}{
 		{
-			// should pick host2, which lacks a ConsumerRef
+			Scenario: "should pick host2, which lacks a ConsumerRef",
 			Machine: machinev1beta1.Machine{
 				ObjectMeta: metav1.ObjectMeta{
 					Name:      "machine1",
@@ -136,7 +202,7 @@ func TestChooseHost(t *testing.T) {
 			Config:           config,
 		},
 		{
-			// should ignore discoveredHost and pick host2, which lacks a ConsumerRef
+			Scenario: "should ignore discoveredHost and pick host2, which lacks a ConsumerRef",
 			Machine: machinev1beta1.Machine{
 				ObjectMeta: metav1.ObjectMeta{
 					Name:      "machine1",
@@ -152,7 +218,7 @@ func TestChooseHost(t *testing.T) {
 			Config:           config,
 		},
 		{
-			// should pick host3, which already has a matching ConsumerRef
+			Scenario: "should pick host3, which already has a matching ConsumerRef",
 			Machine: machinev1beta1.Machine{
 				ObjectMeta: metav1.ObjectMeta{
 					Name:      "machine1",
@@ -171,8 +237,7 @@ func TestChooseHost(t *testing.T) {
 			Config:           config,
 		},
 		{
-			// should not pick a host, because two are already taken, and the third is in
-			// a different namespace
+			Scenario: "should not pick a host, because two are already taken, and the third is in a different namespace",
 			Machine: machinev1beta1.Machine{
 				ObjectMeta: metav1.ObjectMeta{
 					Name:      "machine2",
@@ -191,7 +256,7 @@ func TestChooseHost(t *testing.T) {
 			Config:           config,
 		},
 		{
-			// Can choose hosts with a label, even without a label selector
+			Scenario: "Can choose hosts with a label, even without a label selector",
 			Machine: machinev1beta1.Machine{
 				ObjectMeta: metav1.ObjectMeta{
 					Name:      "machine1",
@@ -205,12 +270,12 @@ func TestChooseHost(t *testing.T) {
 					ProviderSpec: providerSpec,
 				},
 			},
-			Hosts:            []runtime.Object{&host_with_label},
-			ExpectedHostName: host_with_label.Name,
+			Hosts:            []runtime.Object{&hostWithLabel},
+			ExpectedHostName: hostWithLabel.Name,
 			Config:           config,
 		},
 		{
-			// Choose the host with the right label
+			Scenario: "Choose the host with the right label",
 			Machine: machinev1beta1.Machine{
 				ObjectMeta: metav1.ObjectMeta{
 					Name:      "machine1",
@@ -224,12 +289,12 @@ func TestChooseHost(t *testing.T) {
 					ProviderSpec: providerSpec2,
 				},
 			},
-			Hosts:            []runtime.Object{&host2, &host_with_label},
-			ExpectedHostName: host_with_label.Name,
+			Hosts:            []runtime.Object{&host2, &hostWithLabel},
+			ExpectedHostName: hostWithLabel.Name,
 			Config:           config2,
 		},
 		{
-			// No host that matches required label
+			Scenario: "No host that matches required label",
 			Machine: machinev1beta1.Machine{
 				ObjectMeta: metav1.ObjectMeta{
 					Name:      "machine1",
@@ -243,12 +308,12 @@ func TestChooseHost(t *testing.T) {
 					ProviderSpec: providerSpec3,
 				},
 			},
-			Hosts:            []runtime.Object{&host2, &host_with_label},
+			Hosts:            []runtime.Object{&host2, &hostWithLabel},
 			ExpectedHostName: "",
 			Config:           config3,
 		},
 		{
-			// Host that matches a matchExpression
+			Scenario: "Host that matches a matchExpression",
 			Machine: machinev1beta1.Machine{
 				ObjectMeta: metav1.ObjectMeta{
 					Name:      "machine1",
@@ -262,12 +327,12 @@ func TestChooseHost(t *testing.T) {
 					ProviderSpec: providerSpec4,
 				},
 			},
-			Hosts:            []runtime.Object{&host2, &host_with_label},
-			ExpectedHostName: host_with_label.Name,
+			Hosts:            []runtime.Object{&host2, &hostWithLabel},
+			ExpectedHostName: hostWithLabel.Name,
 			Config:           config4,
 		},
 		{
-			// No Host available that matches a matchExpression
+			Scenario: "No Host available that matches a matchExpression",
 			Machine: machinev1beta1.Machine{
 				ObjectMeta: metav1.ObjectMeta{
 					Name:      "machine1",
@@ -286,7 +351,7 @@ func TestChooseHost(t *testing.T) {
 			Config:           config4,
 		},
 		{
-			// No host chosen given an Invalid match expression
+			Scenario: "No host chosen given an Invalid match expression",
 			Machine: machinev1beta1.Machine{
 				ObjectMeta: metav1.ObjectMeta{
 					Name:      "machine1",
@@ -300,41 +365,84 @@ func TestChooseHost(t *testing.T) {
 					ProviderSpec: providerSpec5,
 				},
 			},
-			Hosts:            []runtime.Object{&host2, &host_with_label},
+			Hosts:            []runtime.Object{&host2, &hostWithLabel},
 			ExpectedHostName: "",
 			Config:           config5,
 		},
+		{
+			Scenario: "No host chosen given only externally provisioned choices",
+			Machine: machinev1beta1.Machine{
+				ObjectMeta: metav1.ObjectMeta{
+					Name:      "machine1",
+					Namespace: "myns",
+				},
+				TypeMeta: metav1.TypeMeta{
+					Kind:       "Machine",
+					APIVersion: machinev1beta1.SchemeGroupVersion.String(),
+				},
+				Spec: machinev1beta1.MachineSpec{
+					ProviderSpec: providerSpec5,
+				},
+			},
+			Hosts:            []runtime.Object{&externallyProvisionedHost},
+			ExpectedHostName: "",
+			Config:           config,
+		},
+		{
+			Scenario: "Choose externally provisioned host with consumer ref set",
+			Machine: machinev1beta1.Machine{
+				ObjectMeta: metav1.ObjectMeta{
+					Name:      "machine1",
+					Namespace: "myns",
+				},
+				TypeMeta: metav1.TypeMeta{
+					Kind:       "Machine",
+					APIVersion: machinev1beta1.SchemeGroupVersion.String(),
+				},
+				Spec: machinev1beta1.MachineSpec{
+					ProviderSpec: providerSpec5,
+				},
+			},
+			Hosts: []runtime.Object{&externallyProvisionedHost,
+				&externallyProvisionedAndConsumedHost},
+			ExpectedHostName: "externally-provisioned-and-consumed-host",
+			Config:           config,
+		},
 	}
 
 	for _, tc := range testCases {
-		c := fakeclient.NewFakeClientWithScheme(scheme, tc.Hosts...)
+		t.Run(tc.Scenario, func(t *testing.T) {
+			c := fakeclient.NewFakeClientWithScheme(scheme, tc.Hosts...)
 
-		actuator, err := NewActuator(ActuatorParams{
-			Client: c,
-		})
-		if err != nil {
-			t.Errorf("%v", err)
-		}
-		pspec, err := yaml.Marshal(&tc.Config)
-		if err != nil {
-			t.Logf("could not marshal BareMetalMachineProviderSpec: %v", err)
-			t.FailNow()
-		}
-		tc.Machine.Spec.ProviderSpec = machinev1beta1.ProviderSpec{Value: &runtime.RawExtension{Raw: pspec}}
-		result, err := actuator.chooseHost(context.TODO(), &tc.Machine)
-		if tc.ExpectedHostName == "" {
-			if result != nil {
-				t.Error("found host when none should have been available")
+			actuator, err := NewActuator(ActuatorParams{
+				Client: c,
+			})
+			if err != nil {
+				t.Errorf("%v", err)
 			}
-			continue
-		}
-		if err != nil {
-			t.Errorf("%v", err)
-			return
-		}
-		if result.Name != tc.ExpectedHostName {
-			t.Errorf("host %s chosen instead of %s", result.Name, tc.ExpectedHostName)
-		}
+			pspec, err := yaml.Marshal(&tc.Config)
+			if err != nil {
+				t.Logf("could not marshal BareMetalMachineProviderSpec: %v", err)
+				t.FailNow()
+			}
+			tc.Machine.Spec.ProviderSpec = machinev1beta1.ProviderSpec{Value: &runtime.RawExtension{Raw: pspec}}
+			result, err := actuator.chooseHost(context.TODO(), &tc.Machine)
+			if tc.ExpectedHostName == "" {
+				if result != nil {
+					t.Error("found host when none should have been available")
+				}
+				return
+			}
+			if err != nil {
+				t.Errorf("%v", err)
+				return
+			}
+			if result == nil && tc.ExpectedHostName != "" {
+				t.Errorf("no host found, expected %s", tc.ExpectedHostName)
+			} else if result.Name != tc.ExpectedHostName {
+				t.Errorf("host %s chosen instead of %s", result.Name, tc.ExpectedHostName)
+			}
+		})
 	}
 }
 
@@ -660,18 +768,127 @@ func TestGetHost(t *testing.T) {
 	}
 }
 
+func TestEnsureProviderID(t *testing.T) {
+	scheme := runtime.NewScheme()
+	machinev1beta1.AddToScheme(scheme)
+	bmoapis.AddToScheme(scheme)
+	corev1.AddToScheme(scheme)
+
+	c := fakeclient.NewFakeClientWithScheme(scheme)
+
+	testCases := []struct {
+		Machine machinev1beta1.Machine
+		Host    bmh.BareMetalHost
+		Node    corev1.Node
+	}{
+		{
+			Machine: machinev1beta1.Machine{
+				ObjectMeta: metav1.ObjectMeta{
+					Name:      "mymachine",
+					Namespace: "myns",
+				},
+			},
+			Host: bmh.BareMetalHost{
+				ObjectMeta: metav1.ObjectMeta{
+					Name:      "myhost",
+					Namespace: "myns",
+				},
+			},
+		},
+		{
+			Machine: machinev1beta1.Machine{
+				ObjectMeta: metav1.ObjectMeta{
+					Name:      "mymachine2",
+					Namespace: "myns2",
+				},
+				Status: machinev1beta1.MachineStatus{
+					NodeRef: &corev1.ObjectReference{
+						Kind:      "Node",
+						Namespace: "myns2",
+						Name:      "mynode2",
+					},
+				},
+			},
+			Host: bmh.BareMetalHost{
+				ObjectMeta: metav1.ObjectMeta{
+					Name:      "myhost2",
+					Namespace: "myns2",
+				},
+			},
+			Node: corev1.Node{
+				ObjectMeta: metav1.ObjectMeta{
+					Name:      "mynode2",
+					Namespace: "myns2",
+				},
+			},
+		},
+	}
+
+	for _, tc := range testCases {
+		c.Create(context.TODO(), &tc.Machine)
+		c.Create(context.TODO(), &tc.Host)
+		if tc.Node.Name != "" {
+			c.Create(context.TODO(), &tc.Node)
+		}
+		actuator, err := NewActuator(ActuatorParams{
+			Client: c,
+		})
+		if err != nil {
+			t.Error(err)
+		}
+
+		err = actuator.ensureProviderID(context.TODO(), &tc.Machine, &tc.Host)
+		if err != nil {
+			t.Errorf("unexpected error %v", err)
+		}
+
+		// get the machine and make sure it has the correct ProviderID
+		machine := machinev1beta1.Machine{}
+		key := client.ObjectKey{
+			Name:      tc.Machine.Name,
+			Namespace: tc.Machine.Namespace,
+		}
+		err = c.Get(context.TODO(), key, &machine)
+		providerID := machine.Spec.ProviderID
+		if providerID == nil {
+			t.Error("no machine.Spec.ProviderID found")
+		}
+		expectedProviderID := "baremetalhost:///" + tc.Host.Namespace + "/" + tc.Host.Name
+		if *providerID != expectedProviderID {
+			t.Errorf("ProviderID has value %s, expected %s",
+				*providerID, expectedProviderID)
+		}
+		if tc.Node.Name != "" {
+			node := &corev1.Node{}
+			nodeKey := client.ObjectKey{
+				Name:      tc.Node.Name,
+				Namespace: tc.Node.Namespace,
+			}
+			err = c.Get(context.TODO(), nodeKey, node)
+			nodeProviderID := node.Spec.ProviderID
+			if nodeProviderID != expectedProviderID {
+				t.Errorf("Node ProviderID has value %s, expected %s",
+					nodeProviderID, expectedProviderID)
+			}
+		}
+	}
+}
+
 func TestEnsureAnnotation(t *testing.T) {
 	scheme := runtime.NewScheme()
 	machinev1beta1.AddToScheme(scheme)
 
 	testCases := []struct {
-		Machine machinev1beta1.Machine
-		Host    bmh.BareMetalHost
+		Scenario string
+		Machine  machinev1beta1.Machine
+		Host     bmh.BareMetalHost
 	}{
 		{
-			// annotation exists and is correct
+			Scenario: "annotation exists and is correct",
 			Machine: machinev1beta1.Machine{
 				ObjectMeta: metav1.ObjectMeta{
+					Name:      "mymachine",
+					Namespace: "myns",
 					Annotations: map[string]string{
 						HostAnnotation: "myns/myhost",
 					},
@@ -685,9 +902,11 @@ func TestEnsureAnnotation(t *testing.T) {
 			},
 		},
 		{
-			// annotation exists but is wrong
+			Scenario: "annotation exists but is wrong",
 			Machine: machinev1beta1.Machine{
 				ObjectMeta: metav1.ObjectMeta{
+					Name:      "mymachine",
+					Namespace: "myns",
 					Annotations: map[string]string{
 						HostAnnotation: "myns/wrongvalue",
 					},
@@ -701,9 +920,11 @@ func TestEnsureAnnotation(t *testing.T) {
 			},
 		},
 		{
-			// annotations are empty
+			Scenario: "annotations are empty",
 			Machine: machinev1beta1.Machine{
 				ObjectMeta: metav1.ObjectMeta{
+					Name:        "mymachine",
+					Namespace:   "myns",
 					Annotations: map[string]string{},
 				},
 			},
@@ -715,8 +936,13 @@ func TestEnsureAnnotation(t *testing.T) {
 			},
 		},
 		{
-			// annotations are nil
-			Machine: machinev1beta1.Machine{},
+			Scenario: "annotations are nil",
+			Machine: machinev1beta1.Machine{
+				ObjectMeta: metav1.ObjectMeta{
+					Name:      "mymachine",
+					Namespace: "myns",
+				},
+			},
 			Host: bmh.BareMetalHost{
 				ObjectMeta: metav1.ObjectMeta{
 					Name:      "myhost",
@@ -727,37 +953,39 @@ func TestEnsureAnnotation(t *testing.T) {
 	}
 
 	for _, tc := range testCases {
-		c := fakeclient.NewFakeClientWithScheme(scheme, &tc.Machine)
-		actuator, err := NewActuator(ActuatorParams{
-			Client: c,
+		t.Run(tc.Scenario, func(t *testing.T) {
+			c := fakeclient.NewFakeClientWithScheme(scheme, &tc.Machine)
+			actuator, err := NewActuator(ActuatorParams{
+				Client: c,
+			})
+			if err != nil {
+				t.Error(err)
+			}
+
+			_, err = actuator.ensureAnnotation(context.TODO(), &tc.Machine, &tc.Host)
+			if err != nil {
+				t.Errorf("unexpected error %v", err)
+			}
+
+			// get the machine and make sure it has the correct annotation
+			machine := machinev1beta1.Machine{}
+			key := client.ObjectKey{
+				Name:      tc.Machine.Name,
+				Namespace: tc.Machine.Namespace,
+			}
+			err = c.Get(context.TODO(), key, &machine)
+			annotations := machine.ObjectMeta.GetAnnotations()
+			if annotations == nil {
+				t.Error("no annotations found")
+			}
+			result, ok := annotations[HostAnnotation]
+			if !ok {
+				t.Error("host annotation not found")
+			}
+			if result != "myns/myhost" {
+				t.Errorf("host annotation has value %s, expected \"myns/myhost\"", result)
+			}
 		})
-		if err != nil {
-			t.Error(err)
-		}
-
-		_, err = actuator.ensureAnnotation(context.TODO(), &tc.Machine, &tc.Host)
-		if err != nil {
-			t.Errorf("unexpected error %v", err)
-		}
-
-		// get the machine and make sure it has the correct annotation
-		machine := machinev1beta1.Machine{}
-		key := client.ObjectKey{
-			Name:      tc.Machine.Name,
-			Namespace: tc.Machine.Namespace,
-		}
-		err = c.Get(context.TODO(), key, &machine)
-		annotations := machine.ObjectMeta.GetAnnotations()
-		if annotations == nil {
-			t.Error("no annotations found")
-		}
-		result, ok := annotations[HostAnnotation]
-		if !ok {
-			t.Error("host annotation not found")
-		}
-		if result != "myns/myhost" {
-			t.Errorf("host annotation has value %s, expected \"myns/myhost\"", result)
-		}
 	}
 }
 
@@ -1689,10 +1917,80 @@ func TestDeleteOfBareMetalHostDeletesMachine(t *testing.T) {
 	}
 }
 
+func TestMarshalAndUnmarshal(t *testing.T) {
+	m := map[string]string{"key1": "val1", "keyWithEmptyValue": "", "key.with.dots": "value.with.dots"}
+
+	marshaled, err := marshal(m)
+
+	if err != nil {
+		t.Errorf("marshal() returned an err: %s", err.Error())
+	}
+
+	unmarshaled, err := unmarshal(marshaled)
+
+	if err != nil {
+		t.Errorf("unmarshal returned an err: %s", err.Error())
+	}
+
+	if !reflect.DeepEqual(unmarshaled, m) {
+		t.Errorf("map is different after unmarshal(marshal(map)). Original map was: %s , unmarshaled map: %s",
+			m, unmarshaled)
+	}
+
+	m = nil
+
+	marshaled, err = marshal(m)
+
+	if err != nil {
+		t.Errorf("marshal() returned an err: %s", err.Error())
+	}
+
+	if marshaled != "" {
+		t.Errorf("Expected marshal(nil) to return an empty string, but recieved: %s ", marshaled)
+	}
+
+	unmarshaled, err = unmarshal(marshaled)
+
+	if err != nil {
+		t.Errorf("unmarshal returned an err: %s", err.Error())
+	}
+
+	if len(unmarshaled) != 0 {
+		t.Errorf("Expected unmarshal(marshal(nil)) to return an empty map but recieved: %s", unmarshaled)
+	}
+
+	m = make(map[string]string)
+
+	marshaled, err = marshal(m)
+
+	if err != nil {
+		t.Errorf("marshal() returned an err: %s", err.Error())
+	}
+
+	if marshaled != "{}" {
+		t.Errorf("Expected marshal() with empty map to return {} but got %s", marshaled)
+	}
+
+	unmarshaled, err = unmarshal(marshaled)
+
+	if err != nil {
+		t.Errorf("unmarshal returned an err: %s", err.Error())
+	}
+
+	if len(unmarshaled) != 0 {
+		t.Errorf("Expected unmarshal(marshal([])) to return an empty map but recieved: %s", unmarshaled)
+	}
+}
+
 func TestRemediation(t *testing.T) {
 	machine, machineNamespacedName := getMachine("machine1")
 	host, hostNamespacedName := getBareMetalHost("host1")
 	node, nodeNamespacedName := getNode("node1")
+	nodeAnnotations := map[string]string{"annName1": "annValue1", "annName2": "annValue2"}
+	nodeLabels := map[string]string{"labelName1": "labelValue1", "labelName2": "labelValue2"}
+
+	node.Annotations = nodeAnnotations
+	node.Labels = nodeLabels
 	linkMachineAndNode(machine, node)
 	host.Status.PoweredOn = true
 
@@ -1747,6 +2045,12 @@ func TestRemediation(t *testing.T) {
 		t.Fail()
 	}
 
+	node = &corev1.Node{}
+	err = c.Get(context.TODO(), nodeNamespacedName, node)
+	nodeBackup := node.DeepCopy()
+
+	machine = &machinev1beta1.Machine{}
+	c.Get(context.TODO(), machineNamespacedName, machine)
 	err = actuator.Update(context.TODO(), machine)
 	if err == nil {
 		t.Errorf("expected a requeue err but err was nil")
@@ -1766,11 +2070,30 @@ func TestRemediation(t *testing.T) {
 		t.Fail()
 	}
 
+	//fake client doesn't respect finalizers and deletes the node object
+	//we re-create it with deletion timestamp as this what we expect
+	//from k8s to do
+	now := metav1.Now()
+	nodeBackup.SetDeletionTimestamp(&now)
+
+	nodeBackup.ResourceVersion = "" //create won't work with non empty ResourceVersion
+	c.Create(context.TODO(), nodeBackup)
+
 	err = actuator.Update(context.TODO(), machine)
-	if err != nil {
-		t.Errorf("unexpected error %v", err)
+	expectRequetAfterError(err, t)
+	node = &corev1.Node{}
+	c.Get(context.TODO(), nodeNamespacedName, node)
+
+	if len(node.Finalizers) > 0 {
+		t.Log("Expected finalizer to be removed but finalizer still exist on node")
+		t.Fail()
 	}
 
+	//fake client won't delete the node even though it has no finalizers anymore, so we delete it
+	err = c.Delete(context.TODO(), nodeBackup)
+	err = c.Delete(context.TODO(), node)
+
+	err = actuator.Update(context.TODO(), machine)
 	host = &bmh.BareMetalHost{}
 	c.Get(context.TODO(), hostNamespacedName, host)
 
@@ -1788,11 +2111,31 @@ func TestRemediation(t *testing.T) {
 	c.Create(context.TODO(), node)
 	c.Update(context.TODO(), machine)
 
+	machine = &machinev1beta1.Machine{}
+	c.Get(context.TODO(), machineNamespacedName, machine)
+
 	err = actuator.Update(context.TODO(), machine)
 	if err != nil {
 		t.Errorf("unexpected error %v", err)
 	}
 
+	node = &corev1.Node{}
+	c.Get(context.TODO(), nodeNamespacedName, node)
+
+	assert.Equal(t, nodeAnnotations, node.Annotations,
+		"Node annotations before remediation and after remediation ares not equal")
+
+	assert.Equal(t, nodeLabels, node.Labels,
+		"Node labels before remediation and after remediation ares not equal")
+
+	machine = &machinev1beta1.Machine{}
+	c.Get(context.TODO(), machineNamespacedName, machine)
+	err = actuator.Update(context.TODO(), machine)
+
+	if err != nil {
+		t.Errorf("unexpected error %v", err)
+	}
+
 	machine = &machinev1beta1.Machine{}
 	c.Get(context.TODO(), machineNamespacedName, machine)
 
@@ -1827,6 +2170,19 @@ func TestRemediation(t *testing.T) {
 	}
 }
 
+func expectRequetAfterError(err error, t *testing.T) {
+	if err == nil {
+		t.Errorf("expected a requeue err but err was nil")
+	} else {
+		switch err.(type) {
+		case *machineapierrors.RequeueAfterError:
+			break
+		default:
+			t.Errorf("unexpected error %v", err)
+		}
+	}
+}
+
 func getNode(name string) (*corev1.Node, types.NamespacedName) {
 	node := &corev1.Node{ObjectMeta: metav1.ObjectMeta{
 		Name:      name,
